/*
Name:    BugBox.ino
Created:  5/28/2017 11:05:03 AM
Author: Martin Zvara
*/

#include <VarSpeedServo.h>

/*******************STRUCTS*******************/
typedef struct EXPERIENCE { byte state1; byte state2; float reward; };

/*******************CONSTANTS*******************/
const byte output_size = 12;
//def pins
const int servo1Pin = 11;
const int servo2Pin = 10;
const int servo3Pin = 9;
const int trigPin = 8;
const int echoPin = 7;
const int servo4Pin = 6;

const int minibatch_size = 4;
const int experience_buffer_size = minibatch_size * 2 + 3;

// learning parameters
const float discount_factor = 0.93;
const float learning_rate = 0.00041; 
const float decrease_factor = 0.00094;

/*******************VARIABLES*******************/
byte current_state = 40;
byte previous_state = 40;

byte experience_counter = 0;

int history_distance = 0;
int last_distance = 0;

long duration = 0;

float rand_action_prob = 1.35;

float new_q_vals[output_size] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
float old_q_vals[output_size] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

float input[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 1 }; //input current, previous action
float hidden1[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 1 };
float hidden2[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 1 };
float output[12] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

float in_hid1_weights[9][8] = 
{ { 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 } };
float hid1_hid2_weights[9][8] = 
{ { 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0 } };
float hid2_out_weights[9][12] =
{ { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } };

EXPERIENCE experience_buffer[experience_buffer_size];
VarSpeedServo servo[4];

/*******************HARDWARE I/O FUNCTIONS*******************/
int GetDistance()
{
  int tmp;
  int dists[5];

  for (byte i = 0; i < 5; i++)
  {
    // Clears the trigPin
    digitalWrite(trigPin, LOW);
    delayMicroseconds(2);

    // Sets the trigPin on HIGH state for 10 micro seconds
    digitalWrite(trigPin, HIGH);
    delayMicroseconds(10);
    digitalWrite(trigPin, LOW);

    // Reads the echoPin, returns the sound wave travel time in microseconds
    duration = pulseIn(echoPin, HIGH);

    // Calculate and save the distance to array
    dists[i] = ((duration / 2) / 29.1);
  }

  // sort array of distances
  for (byte i = 0; i < 4; i++) {
    for (byte j = 0; j < (5 - (i + 1)); j++) {
      if (dists[j] > dists[j + 1]) {
        tmp = dists[j];
        dists[j] = dists[j + 1];
        dists[j + 1] = tmp;
      }
    }
  }

  //return median
  return dists[2];
}

/**
* Execute action and wait until all servo motors are in right position.
* Params:
* float action[]
*/
float ExecuteAction(float action[])
{
  byte max_rotation;
  byte max_index = 0;
  byte diff = -10;
  for (byte i = 0; i < 4; i++)
  {
    max_rotation = abs(GetAngleFromAction(a[i]) - servo[i].read());
    if (diff <= max_rotation)
    {
      diff = max_rotation;
      max_index = i;
    }
  }

  for (byte i = 0; i < 4; i++)
  {
    if (i != max_index)
    {
      servo[i].write(GetAngleFromAction(a[i]), 25);
    }
  }
  servo[max_index].write(GetAngleFromAction(a[max_index]), 25, true);
}

/*******************DQN FUNCTIONS*******************/
float Tanh(float x)
{
   return 2. / (1. + exp(-2. * x)) - 1.;
}

/**
* Feed forward an action a1, a2 through our neural net.
* Params:
* float a1[] -> current action(array of four pasition for each servo)
* float a2[] -> previous action(array of four pasition for each servo)
*/
void FeedNet(float a1[], float a2[])
{
  //feed inputs
  for (byte i = 0; i < 4; i++)
  {
    if (i == 1 || i == 2)
    {
      input[i] = (2-a1[i]) * 0.5;
    }
    else
    {
      input[i] = a1[i] * 0.5;
    }
  }

  for (byte i = 0; i < 4; i++)
  {
    if (i == 1 || i == 2)
    {
      input[i + 4] = (2 - a2[i]) * 0.5;
    }
    else
    {
      input[i + 4] = a2[i] * 0.5;
    }
  }

  //at first, clear values on hidden1 layer from previous feed
  for (byte i = 0; i < 8; i++)
  {
    hidden1[i] = 0.;
  }
  //continue to hidden layer 
  for (byte i = 0; i < 8; i++)
  {
    for (byte j = 0; j < 9; j++)
    {
      hidden1[i] += input[j] * in_hid1_weights[j][i];
    }
  }
  //activate hidden perceptrons
  for (byte i = 0; i < 8; i++)
  {
    hidden1[i] = Tanh(hidden1[i]);
  }

  //at first, clear values on hidden2 layer from previous feed
  for (byte i = 0; i < 8; i++)
  {
    hidden2[i] = 0.;
  }
  //continue to hidden layer 
  for (byte i = 0; i < 8; i++)
  {
    for (byte j = 0; j < 9; j++)
    {
      hidden2[i] += hidden1[j] * hid1_hid2_weights[j][i];
    }
  }
  //activate hidden perceptrons
  for (byte i = 0; i < 8; i++)
  {
    hidden2[i] = Tanh(hidden2[i]);
  }

  //finally, clear and pass everithing to output
  for (byte i = 0; i < 12; i++)
  {
    output[i] = 0.;
  }
  for (byte i = 0; i < 12; i++)
  {
    for (byte j = 0; j < 9; j++)
    {
      output[i] += hidden2[j] * hid2_out_weights[j][i];
    }
  }
  //activate output perceptron

  for (byte i = 0; i < 12; i++)
  {
    output[i] = Tanh(output[i]);
  }
}

/**
* Adjust weights of QNetwork according error. Error is calculated according value iteration of q-learning:
* 1/2*(reward + learningRate * argmax[a']Q(s',a') - Q(s,a)).
* Params:
* float maxQ[] -> target value (argmax[a']Q(s',a'))
* float reward
*/
void BackpropagateNet(float maxQ[], float reward)
{
  float delta_output[12] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float delta_hid2[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
  float delta_hid1[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

  for (byte i = 0; i < 12; i++)
  {
    delta_output[i] = (((reward + (discount_factor * maxQ[i])) - output[i])) * (1 - output[i] * output[i]);
  }

  for (byte i = 0; i < 8; i++)
  {
    for (byte j = 0; j < 12; j++)
    {
      delta_hid2[i] += delta_output[j] * hid2_out_weights[i][j];
    }
  }

  for (byte i = 0; i < 8; i++)
  {
    for (byte j = 0; j < 8; j++)
    {
      delta_hid1[i] += delta_hid2[j] * hid1_hid2_weights[i][j];
    }
  }

  for (byte i = 0; i < 9; i++)
  {
    for (byte j = 0; j < 8; j++)
    {
      in_hid1_weights[i][j] = in_hid1_weights[i][j] - (learning_rate * (1 - hidden1[j] * hidden1[j]) * delta_hid1[j] * input[i]);
    }
  }

  for (byte i = 0; i < 9; i++)
  {
    for (byte j = 0; j < 8; j++)
    {
      hid1_hid2_weights[i][j] = hid1_hid2_weights[i][j] - (learning_rate * (1 - hidden2[j] * hidden2[j]) * delta_hid2[j] * hidden1[i]);
    }
  }

  for (byte i = 0; i < 9; i++)
  {
    for (byte j = 0; j < 12; j++)
    {
      hid2_out_weights[i][j] = hid2_out_weights[i][j] - (learning_rate * (delta_output[j] * hidden2[i]));
    }
  }
}


/*******************UTILITIES*******************/
void GetAction(int state, float action[])
{
  int a = state;

  for (byte i = 3; i > 0; i--)
  {
    action[i] = (a % 3);
    a /= 3;
  }
  action[0] = (a % 3);
}

int GetState(float action[])
{
  int a = 0;
  for (byte i = 0; i < 4; i++)
  {
    a += (action[i] * (int)pow(3, 3 - i));
  }
  return a;
}

byte GetAngleFromAction(float a)
{
  if (a == 0)
  {
    return 70;
  }
  else if (a == 1)
  {
    return 90;
  }
  else if (a == 2)
  {
    return 110;
  }
}

unsigned int bitOut(void)
{
  static unsigned long firstTime = 1, prev = 0;
  unsigned long bit1 = 0, bit0 = 0, x = 0, port = 0, limit = 99;
  if (firstTime)
  {
    firstTime = 0;
    prev = analogRead(port);
  }
  while (limit--)
  {
    x = analogRead(port);
    bit1 = (prev != x ? 1 : 0);
    prev = x;
    x = analogRead(port);
    bit0 = (prev != x ? 1 : 0);
    prev = x;
    if (bit1 != bit0)
      break;
  }
  return bit1;
}
//------------------------------------------------------------------------------
unsigned long seedOut(unsigned int noOfBits)
{
  // return value with 'noOfBits' random bits set
  unsigned long seed = 0;
  for (int i = 0; i < noOfBits; ++i)
    seed = (seed << 1) | bitOut();
  return seed;
}

void swapQVals(float q1[], float q2[], int lenght)
{
  float tmp;
  for (byte i = 0; i < lenght; i++)
  {
    tmp = q1[i];
    q1[i] = q2[i];
    q2[i] = tmp;
  }
}

/*******************PHASE FUNCTIONS*******************/

/**
* This function will execute one new transitions (s1, s2, r) where s2 is chosen acording epsilon greedy 
* strategy and add it to experience buffer
*/
void ExploreActions()
{
  byte new_state = current_state;

  int distance = 0;

  float reward = -0.2;
  float action1[4] = { 0, 0, 0, 0 };
  float action2[4] = { 0, 0, 0, 0 };

  if ((float)random(1, 101) / 100.0 < rand_action_prob) //epsilon greedy strategy
  {
    while (new_state == current_state)
    {
      new_state = random(0, 81);
    }
    GetAction(new_state, action1);
    Serial.println("R");
  }
  else
  {
    GetAction(previous_state, action1);
    GetAction(current_state, action2);
    FeedNet(action2, action1);
    for (byte i = 0; i < 4; i++)
    {
      GetMaxQandAction(i, action1);
    }
    new_state = GetState(action1);
    Serial.println("M");
  }

  rand_action_prob = rand_action_prob > decrease_factor ? rand_action_prob - decrease_factor : 0;

  ExecuteAction(action1);
  delay(50);
  distance = GetDistance();

  if (distance - last_distance < -1)
  {
    reward = 0.3;
  }
  else if (distance - last_distance > 1)
  {
    reward = -0.3;
  }
  if (new_state != current_state) 
  {
    for (byte i = 0; i < 4; i++)
    {
      if (action1[i] != action2[i])
      {
        reward += 0.1;
      }
    } 
  }
  if (history_distance > distance)
  {
    reward += 0.6;
  }
  else
  {
    reward -= 0.6;
  }
  experience_buffer[experience_counter].state1 = current_state;
  experience_buffer[experience_counter].state2 = new_state;
  experience_buffer[experience_counter].reward = reward;

  //Serial.print("Reward: ");
  Serial.println(reward);

  history_distance = last_distance;
  last_distance = distance;
  previous_state = current_state;
  current_state = new_state;
  experience_counter += 1;
}

float GetMaxQandAction(byte servo, float* action)
{
  float q_values[3] = {input[servo], input[servo + 4], input[servo + 8]};
  float max_q = -10.0;
  float tmp = 2;
  for (byte q = 0; q < 3; q++)
  {
    if (max_q < q_values[q])
    {
      
      if (servo == 1 || servo == 2)
      {
        action[servo] = tmp - q;
      }
      else
      {
        action[servo] = q;
      }
      max_q = q_values[q];
    }
  }
  return max_q;
}

void setup()
{
  Serial.begin(9600);

  //attach servos to coresponding pins
  servo[0].attach(servo1Pin);
  servo[1].attach(servo2Pin);
  servo[2].attach(servo3Pin);
  servo[3].attach(servo4Pin);

  //set default position and state
  servo[0].write(90, 200);
  servo[1].write(90, 200);
  servo[2].write(90, 200);
  servo[3].write(90, 200, true);

  //set trigger and echo pins
  pinMode(trigPin, OUTPUT);
  pinMode(echoPin, INPUT);

  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, HIGH);

  //seed random
  randomSeed(seedOut(31));

  //from Glorot and Bengio Understanding the difficulty of training deep feedforward neural networks
  //for tanh init weight uniformly from interval  +- sqrt(6 / (current_layer_size + next_layer_size))
  for (byte i = 0; i < 9; i++)
  {
    for (byte j = 0; j < 8; j++)
    {
      in_hid1_weights[i][j] = (float)random(0, 59) / 100.0;
      if (random(0, 101) > 50)
      {
        in_hid1_weights[i][j] *= -1;
      }
    }
  }
  for (byte i = 0; i < 9; i++)
  {
    for (byte j = 0; j < 8; j++)
    {
      hid1_hid2_weights[i][j] = (float)random(0, 59) / 100.0;
      if (random(0, 101) > 50)
      {
        hid1_hid2_weights[i][j] *= -1;
      }
    }
  }
  for (byte i = 0; i < 9; i++)
  {
    for (byte j = 0; j < 12; j++)
    {
      hid2_out_weights[i][j] = (float)random(0, 54) / 100.0;
      if (random(0, 101) > 50)
      {
        hid2_out_weights[i][j] *= -1;
      }
    }
  }
  for (byte i = 0; i < 12; i++)
  {
    new_q_vals[i] = ((float)random(0, 20) / 100.0) * (random(0, 101) > 50 ? -1 : 1);
  }
  history_distance = GetDistance();
  last_distance = history_distance;

  delay(3000);

  digitalWrite(LED_BUILTIN, LOW);
}

void loop()
{
  if (experience_counter < experience_buffer_size) //exploring action
  {
    Serial.println("Acting");
    ExploreActions();
  }
  else //teach explored actions
  {
    float action1[4];
    float action2[4];

    for (byte i = 0; i < experience_buffer_size; i++)
    {
      Serial.println("Teaching");
    
      swapQVals(new_q_vals, old_q_vals, output_size);

      GetAction(experience_buffer[i+1].state1, action1);
      GetAction(experience_buffer[i+1].state2, action2);
      FeedNet(action2, action1);

      for (byte j = 0; j < output_size; j++)
      {
        new_q_vals[j] = output[j];
      }

      GetAction(experience_buffer[i].state1, action1);
      GetAction(experience_buffer[i].state2, action2);
      FeedNet(action2, action1);

      BackpropagateNet(old_q_vals, experience_buffer[i].reward);
    }
    experience_counter = 0;
  }
}



